
# django_auto_url - Automagic URLs for Django
# Copyright (C) 2019 Thomas Hartmann <thomas.hartmann@th-ht.de>
#
# This file is part of django_auto_url.
#
# django_auto_url is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# django_auto_url is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with django_auto_url.  If not, see <http://www.gnu.org/licenses/>.

from django.views.generic import TemplateView, DetailView

from django_auto_url import kwargs
from django_auto_url.mixins import AutoUrlMixin
from django_auto_url.urls import reverse_local_classname
from django_auto_url.tests import models


class EmptyView(AutoUrlMixin, TemplateView):
    """Most generic Test View."""

    template_name = 'generic.html'


class IndexView(AutoUrlMixin, TemplateView):
    """View to test is_index."""

    is_index = True
    template_name = 'index.html'


class KwargsView(AutoUrlMixin, TemplateView):
    """View to test Kwargs."""

    template_name = 'use_kwargs.html'
    url_kwargs = [kwargs.string('string'),
                  kwargs.int('int'),
                  kwargs.bool('bool')]


class KwargsViewWithDefaults(AutoUrlMixin, TemplateView):
    """View to test Kwargs with defaults."""

    template_name = 'use_kwargs.html'
    url_kwargs = [kwargs.string('string'),
                  kwargs.int('int', 33),
                  kwargs.bool('bool', True)]


class WithLinksView(AutoUrlMixin, TemplateView):
    """View that has a template with links."""

    template_name = 'with_links.html'


class WithInvalidArgInTemplateTagView(AutoUrlMixin,
                                      TemplateView):
    """View that has a template with invalid links."""

    template_name = 'with_args_link.html'


class WithLinkTargetFromContextView(AutoUrlMixin,
                                    TemplateView):
    """View that has a link defining the target from the context."""

    template_name = 'link_target_from_context.html'
    extra_context = {
        'link_target': reverse_local_classname('EmptyView')
    }


class StringItemFromPK(AutoUrlMixin, DetailView):
    """View displaying a StringItem."""

    template_name = 'string_object.html'
    model = models.StringItem


class StringItemFromCustomPK(AutoUrlMixin, DetailView):
    """View displaying a StringItem using custom PK field."""

    template_name = 'string_object.html'
    model = models.StringItem
    pk_url_kwarg = 'test_pk'


class StringItemFromSlug(AutoUrlMixin, DetailView):
    """View displaying a StringItem using the slug field."""

    template_name = 'string_object.html'
    model = models.StringItem
    pk_url_kwarg = None
    slug_field = 'test_slug'




# django_auto_url - Automagic URLs for Django
# Copyright (C) 2019 Thomas Hartmann <thomas.hartmann@th-ht.de>
#
# This file is part of django_auto_url.
#
# django_auto_url is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# django_auto_url is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with django_auto_url.  If not, see <http://www.gnu.org/licenses/>.

# This is a custom version of the lazy implementation taken from Django.
# It returns the type of the wrapped return type correctly.

from functools import total_ordering, wraps

import six
from django.utils.functional import _lazy_proxy_unpickle


class Promise(metaclass=type(str)):
    """
    This is just a base class for the proxy class created in
    the closure of the lazy function. It can be used to recognize
    promises in code.
    """
    pass


def lazy(func, *resultclasses):
    """
    Turns any callable into a lazy evaluated callable. You need to give result
    classes or types -- at least one is needed so that the automatic forcing of
    the lazy evaluation code is triggered. Results are not memoized; the
    function is evaluated on every access.
    """

    @total_ordering
    class __proxy__(Promise):
        """
        Encapsulate a function call and act as a proxy for methods that are
        called on the result of that function. The function is not evaluated
        until one of the methods on the result is called.
        """
        __prepared = False

        def __init__(self, args, kw):
            super().__init__()
            self.__args = args
            self.__kw = kw
            if not self.__prepared:
                self.__prepare_class__()
            self.__prepared = True

        def __reduce__(self):
            return (
                _lazy_proxy_unpickle,
                (func, self.__args, self.__kw) + resultclasses
            )

        def __repr__(self):
            return repr(self.__cast())

        @classmethod
        def __prepare_class__(cls):
            for resultclass in resultclasses:
                for type_ in resultclass.mro():
                    for method_name in type_.__dict__.keys():
                        # All __promise__ return the same wrapper method, they
                        # look up the correct implementation when called.
                        if hasattr(cls, method_name):
                            continue
                        meth = cls.__promise__(method_name)
                        setattr(cls, method_name, meth)
            cls._delegate_bytes = bytes in resultclasses
            cls._delegate_text = six.text_type in resultclasses
            assert not (cls._delegate_bytes and cls._delegate_text), (
                "Cannot call lazy() with both bytes and text return types.")
            if cls._delegate_text:
                if six.PY3:
                    cls.__str__ = cls.__text_cast
                else:
                    cls.__unicode__ = cls.__text_cast
                    cls.__str__ = cls.__bytes_cast_encoded
            elif cls._delegate_bytes:
                if six.PY3:
                    cls.__bytes__ = cls.__bytes_cast
                else:
                    cls.__str__ = cls.__bytes_cast

        @classmethod
        def __promise__(cls, method_name):
            # Builds a wrapper around some magic method
            def __wrapper__(self, *args, **kw):
                # Automatically triggers the evaluation of a lazy value and
                # applies the given magic method of the result type.
                res = func(*self.__args, **self.__kw)
                return getattr(res, method_name)(*args, **kw)

            return __wrapper__

        def __text_cast(self):
            return func(*self.__args, **self.__kw)

        def __bytes_cast(self):
            return bytes(func(*self.__args, **self.__kw))

        def __bytes_cast_encoded(self):
            return func(*self.__args, **self.__kw).encode('utf-8')

        def __cast(self):
            if self._delegate_bytes:
                return self.__bytes_cast()
            elif self._delegate_text:
                return self.__text_cast()
            else:
                return func(*self.__args, **self.__kw)

        def __str__(self):
            # object defines __str__(), so __prepare_class__() won't overload
            # a __str__() method from the proxied class.
            return str(self.__cast())

        def __ne__(self, other):
            if isinstance(other, Promise):
                other = other.__cast()
            return self.__cast() != other

        def __eq__(self, other):
            if isinstance(other, Promise):
                other = other.__cast()
            return self.__cast() == other

        def __lt__(self, other):
            if isinstance(other, Promise):
                other = other.__cast()
            return self.__cast() < other

        def __hash__(self):
            return hash(self.__cast())

        def __mod__(self, rhs):
            if self._delegate_bytes and six.PY2:
                return bytes(self) % rhs
            elif self._delegate_text:
                return six.text_type(self) % rhs
            return self.__cast() % rhs

        def __deepcopy__(self, memo):
            # Instances of this class are effectively immutable. It's just a
            # collection of functions. So we don't need to do anything
            # complicated for copying.
            memo[id(self)] = self
            return self

        def __getattribute__(self, item):
            if item == '__class__':
                return resultclasses[0]
            else:
                return super().__getattribute__(item)

    @wraps(func)
    def __wrapper__(*args, **kw):
        # Creates the proxy object, instead of the actual value.
        return __proxy__(args, kw)

    return __wrapper__




# django_auto_url - Automagic URLs for Django
# Copyright (C) 2019 Thomas Hartmann <thomas.hartmann@th-ht.de>
#
# This file is part of django_auto_url.
#
# django_auto_url is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# django_auto_url is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with django_auto_url.  If not, see <http://www.gnu.org/licenses/>.


import pytest
import seleniumbase
from django_auto_url.utils.class2urlname import Class2URLName


@pytest.fixture(scope='session', autouse=True)
def init_tests(tmpdir_factory):
    """General init fixture for all pytests."""
    # settings.STATIC_ROOT = str(tmpdir_factory.mktemp('static'))
    # call_command('collectstatic', interactive=False)
    yield


@pytest.fixture(scope='module', autouse=True)
def init_tests_module():
    Class2URLName.url_name_cache = None
    yield


@pytest.fixture(scope='module')
def selenium():
    """Fixture returning a seleniumbase."""
    this_sel = seleniumbase.BaseCase('__init__')
    this_sel.setUp()
    yield this_sel
    this_sel.tearDown()




# django_auto_url - Automagic URLs for Django
# Copyright (C) 2019 Thomas Hartmann <thomas.hartmann@th-ht.de>
#
# This file is part of django_auto_url.
#
# django_auto_url is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# django_auto_url is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with django_auto_url.  If not, see <http://www.gnu.org/licenses/>.
import pytest

from django_auto_url.urls import reverse_classname, reverse_classname_lazy
from django_auto_url.tests.views import test_views
from django_auto_url.tests.factories import StringItemFactory
from django_auto_url.tests.utils import assert_stringitem


def test_simple_access(selenium, live_server):
    """Simple test for reverse_classname."""
    selenium.open(live_server.url +
                  reverse_classname(
                      test_views.EmptyView))

    selenium.assert_text('Hello')


def test_simple_access_lazy(selenium, live_server):
    """Simple test for reverse_classname_lazy."""
    selenium.open(live_server.url +
                  str(reverse_classname_lazy(
                      test_views.EmptyView)))

    selenium.assert_text('Hello')


def test_index(selenium, live_server):
    """Test index view."""
    selenium.open(live_server.url)
    selenium.assert_text('Index')


def test_kwargs(selenium, live_server):
    """Test with kwargs."""
    my_int = 42
    my_string = 'Oh Hello!'
    my_bool = True

    selenium.open(live_server +
                  reverse_classname(
                      test_views.KwargsView,
                      kwargs={'int': my_int,
                              'string': my_string,
                              'bool': my_bool}
                  ))

    with pytest.raises(RuntimeError, match='args and kwargs not allowed if '
                                           'return_url_name=True.'):
        reverse_classname(
            test_views.KwargsView,
            return_url_name=True,
            kwargs={'int': my_int,
                    'string': my_string,
                    'bool': my_bool}
        )

    selenium.assert_text('Int: ' + str(my_int))
    selenium.assert_text('String: ' + my_string)
    selenium.assert_text('Bool: ' + str(my_bool))

    my_int = 22
    my_string = 'Blabla'
    my_bool = False

    selenium.open(live_server +
                  reverse_classname(
                      test_views.KwargsView,
                      args=[my_string,
                            my_int,
                            my_bool]
                  ))

    selenium.assert_text('Int: ' + str(my_int))
    selenium.assert_text('String: ' + my_string)
    selenium.assert_text('Bool: ' + str(my_bool))

    with pytest.raises(RuntimeError, match='args and kwargs not allowed if '
                                           'return_url_name=True.'):
        reverse_classname(
            test_views.KwargsView,
            return_url_name=True,
            args=[my_string,
                  my_int,
                  my_bool]
        )


def test_links(selenium, live_server):
    """Test with links in template."""
    selenium.open(live_server +
                  reverse_classname(
                      test_views.WithLinksView)
                  )
    selenium.click('#index_link')
    selenium.assert_text('Index')
    selenium.open(live_server +
                  reverse_classname(
                      test_views.WithLinksView)
                  )
    selenium.click('#kwargs_link')
    selenium.assert_text('Int: ' + str(32))
    selenium.assert_text('String: ' + 'New String')
    selenium.assert_text('Bool: ' + str(False))

    selenium.open(live_server +
                  reverse_classname(
                      test_views.WithLinksView)
                  )
    selenium.click('#kwargs_default_link')
    selenium.assert_text('Int: ' + str(33))
    selenium.assert_text('String: ' + 'New New String')
    selenium.assert_text('Bool: ' + str(True))


def test_kwargs_with_defaults(selenium, live_server):
    """Test with default kwargs."""
    my_int = 22
    my_string = 'Blabla'
    my_bool = False

    default_int = 33
    default_bool = True

    selenium.open(live_server +
                  reverse_classname(
                      test_views.KwargsViewWithDefaults,
                      kwargs={'int': my_int,
                              'string': my_string,
                              'bool': my_bool}
                  ))

    selenium.assert_text('Int: ' + str(my_int))
    selenium.assert_text('String: ' + my_string)
    selenium.assert_text('Bool: ' + str(my_bool))

    selenium.open(live_server + reverse_classname(
        test_views.KwargsViewWithDefaults,
        kwargs={'string': my_string}
    ))

    selenium.assert_text('Int: ' + str(default_int))
    selenium.assert_text('String: ' + my_string)
    selenium.assert_text('Bool: ' + str(default_bool))

    selenium.open(live_server + reverse_classname(
        test_views.KwargsViewWithDefaults,
        kwargs={'int': my_int,
                'string': my_string}
    ))

    selenium.assert_text('Int: ' + str(my_int))
    selenium.assert_text('String: ' + my_string)
    selenium.assert_text('Bool: ' + str(default_bool))


def test_view_with_args_templatetag(selenium, live_server, settings):
    """Test if args in templatetag are forbidden."""
    settings.DEBUG = True
    selenium.open(live_server +
                  reverse_classname(
                      test_views.WithInvalidArgInTemplateTagView))

    selenium.assert_text('Only keyword arguments are '
                         'allowed when using the '
                         '"url_from_class" template tag.')


def test_access_suburl(selenium, live_server):
    """Test whether suburls work."""
    selenium.open(live_server + reverse_classname(
        'django_auto_url.tests.views.test_sub_views.NormalSubView'
    ))

    selenium.assert_text('Hello')


def test_local_reverse(selenium, live_server):
    """Test local reverse."""
    selenium.open(live_server + reverse_classname(
        test_views.WithLinkTargetFromContextView
    ))

    selenium.click('#link')

    selenium.assert_text('Hello')


@pytest.mark.django_db
def test_pk_slug(selenium, live_server):
    """Test pk and slug."""
    all_items = StringItemFactory.create_batch(10)

    selenium.open(live_server.url)
    for cur_item in all_items:
        selenium.open(live_server +
                      reverse_classname(test_views.StringItemFromPK,
                                        kwargs={'pk': cur_item.pk}))

        assert_stringitem(selenium, cur_item)

        selenium.open(live_server +
                      reverse_classname(test_views.StringItemFromCustomPK,
                                        kwargs={'test_pk': cur_item.pk}))

        assert_stringitem(selenium, cur_item)

        selenium.open(live_server +
                      reverse_classname(test_views.StringItemFromSlug,
                                        kwargs={'slug': cur_item.test_slug}))

        assert_stringitem(selenium, cur_item)




# django_auto_url - Automagic URLs for Django
# Copyright (C) 2019 Thomas Hartmann <thomas.hartmann@th-ht.de>
#
# This file is part of django_auto_url.
#
# django_auto_url is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# django_auto_url is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with django_auto_url.  If not, see <http://www.gnu.org/licenses/>.

import inspect

import django
import django.urls
from django.urls import URLPattern, URLResolver
lookup_str_imported = False

try:
    from django.contrib.admindocs.utils import lookup_str
    lookup_str_imported = True
except ImportError:
    pass

url_cache = {}


class Class2URLName(object):
    """Keeps the url_name_cache, scans for views, does the lookups."""

    url_name_cache = None

    @classmethod
    def get_lookup_class_urlname(cls, class_name):
        """Convert a class or class name into a url name."""
        if not cls.url_name_cache:
            cls.generate_url_names_cache()

        if isinstance(class_name, type):
            class_name = '.'.join(
                [inspect.getmodule(class_name).__name__, class_name.__name__])

        if not isinstance(class_name, str):
            raise ValueError('You must supply either a string or a class')

        try:
            url = cls.url_name_cache[class_name]
        except KeyError:
            raise ValueError('Supplied class name "%s" does not match any '
                             'view classes' % (class_name,))

        return url

    @classmethod
    def generate_url_names_cache(cls):
        """Generate the url_name_cache."""
        cls.url_name_cache = cls.parse_urls(
            django.urls.get_resolver().url_patterns)

    @classmethod
    def parse_urls(cls, patterns):
        """Look for urlpatterns for our views."""
        all_patterns = {}
        if isinstance(patterns, list):
            for cur_pattern in patterns:
                all_patterns.update(cls.parse_urls(cur_pattern))

        if isinstance(patterns, URLResolver):
            cur_namespace = patterns.namespace
            all_items = {}
            for cur_pattern in patterns.url_patterns:
                all_items.update(cls.parse_urls(cur_pattern))

            for cur_key, cur_item in all_items.items():
                if not cur_namespace:
                    all_items[cur_key] = '%s' % (cur_item,)
                else:
                    all_items[cur_key] = '%s:%s' % (cur_namespace, cur_item)

            return all_items

        if isinstance(patterns, URLPattern):
            if lookup_str_imported:
                key = lookup_str(patterns)
            else:
                key = patterns.lookup_str

            return {key: patterns.name}

        return all_patterns




# django_auto_url - Automagic URLs for Django
# Copyright (C) 2019 Thomas Hartmann <thomas.hartmann@th-ht.de>
#
# This file is part of django_auto_url.
#
# django_auto_url is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# django_auto_url is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with django_auto_url.  If not, see <http://www.gnu.org/licenses/>.

from django.urls import re_path

from django_auto_url import kwargs


class AutoUrlMixin:
    """Mixin for automatic URL url creation.

    Include this Mixin in your :class:`~django.views.generic.base.View`
    to have a urlpattern automatically
    generated when this surrounding package is scanned with
    :func:`~django_auto_url.urls.get_urls_from_module`.

    Attributes
    ----------
    url_path_name : str, optional
        If set, this is the string that appears in the actual URL. Otherwise
        the class name is used.

    url_django_name : str, optional
        If set, this is the name of the route in django. Otherwise
        the class name is used.

    is_index : bool
        Set to :code:`True` if this is an index view.

    url_kwargs : list of :class:`~django_auto_url.kwargs.kwargs.kwarg`
        The list of Keyword Arguments for this view.

    url_ignore_pk : bool
        If the :class:`~django.views.generic.base.View` includes a
        :code:`slug_url_kwarg` or :code:`pk_url_kwarg` class attribute
        (like the :class:`~django.views.generic.detail.DetailView`),
        a keyword argument for the Primary Key or the Slug is included
        automatically. Set this to :code:`False` if this is not what you want.
    """

    url_path_name = None
    url_kwargs = None
    url_django_name = None
    is_index = False
    url_ignore_pk = False

    @classmethod
    def get_url_path_name(cls):
        """Return the url_path_name."""
        if cls.url_path_name:
            return cls.url_path_name
        else:
            return cls.__name__

    @classmethod
    def get_url_kwargs(cls):
        """Return the Keyword Arguments."""
        url_kwargs = cls.url_kwargs
        if not url_kwargs and hasattr(cls, 'pk_url_kwarg') \
                and not cls.url_ignore_pk and cls.pk_url_kwarg is not None:
            if cls.pk_url_kwarg == 'pk':
                url_kwargs = [kwargs.int('pk')]
            else:
                url_kwargs = [
                    kwargs.string(cls.pk_url_kwarg)]

        if not url_kwargs and hasattr(cls, 'slug_url_kwarg') \
                and not cls.url_ignore_pk and cls.slug_url_kwarg is not None:
            url_kwargs = [
                kwargs.string(cls.slug_url_kwarg)]

        return url_kwargs

    @classmethod
    def get_url_django_name(cls):
        """Return the django url name."""
        if cls.url_django_name:
            return cls.url_django_name
        else:
            return cls.__name__

    @classmethod
    def get_url(cls):
        """
        Generate and return the :class:`~django.urls.resolvers.URLResolver`s.

        Returns
        -------
        list of :class:`~django.urls.resolvers.URLResolver`
        """
        if cls.is_index and cls.url_kwargs:
            raise RuntimeError('An Index View cannot have kwargs.')

        if cls.is_index:
            url_part = r'^$'
            return [re_path(url_part, cls.as_view(),
                            name=cls.get_url_django_name())]
        else:
            return cls._create_url(cls.get_url_path_name(),
                                   cls.get_url_kwargs(),
                                   cls.get_url_django_name())

    @classmethod
    def _create_url(cls, path_name, url_kwargs, django_name):
        url_list = []

        url_part = r'^%s/' % (path_name,)

        if url_kwargs:
            if url_kwargs[-1].default is not None:
                new_url_kwargs = url_kwargs[0:-1]
                url_list = cls._create_url(
                    path_name, new_url_kwargs, django_name) + url_list

            for cur_kwarg in url_kwargs:
                url_part = url_part + r'%s/' % (cur_kwarg.get_regex())

        url_part = url_part + r'$'

        url_list.insert(0, re_path(url_part, cls.as_view(), name=django_name))

        return url_list

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def dispatch(self, request, *args, **kwargs):  # noqa: D102
        total_n_args = len(args) + len(kwargs)
        if self.url_kwargs:
            for idx_kwarg, cur_kwarg in enumerate(self.url_kwargs):
                if idx_kwarg >= total_n_args and \
                        cur_kwarg.name not in kwargs:
                    kwargs[cur_kwarg.name] = cur_kwarg.default

        return super().dispatch(request, *args, **kwargs)




# django_auto_url - Automagic URLs for Django
# Copyright (C) 2019 Thomas Hartmann <thomas.hartmann@th-ht.de>
#
# This file is part of django_auto_url.
#
# django_auto_url is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# django_auto_url is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with django_auto_url.  If not, see <http://www.gnu.org/licenses/>.

from django_auto_url.urls import get_urls_from_module
from django_auto_url.tests.views import test_bad_index_view

app_name = 'test_for_autourl'

urlpatterns = get_urls_from_module(test_bad_index_view)




# django_auto_url - Automagic URLs for Django
# Copyright (C) 2019 Thomas Hartmann <thomas.hartmann@th-ht.de>
#
# This file is part of django_auto_url.
#
# django_auto_url is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# django_auto_url is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with django_auto_url.  If not, see <http://www.gnu.org/licenses/>.


class kwarg(object):
    """Base class for Keyword Arguments.

    Attributes
    ----------
    regex : str
        The regular expression for this type.

    name : str
        The name of this keyword argument.

    default
        The default value for this keyword argument. None if no default.
    """

    regex = None

    def __init__(self, name, default=None):
        self.name = name
        self.default = default

    def get_regex(self):
        """Return the regular expression for this Keyword Argument.

        Returns
        -------
        str
            The generated regular expression.
        """
        return r'(?P<%s>%s)' % (self.name, self.regex)


class string(kwarg):
    """String Keyword Argument."""

    regex = '.*'


class int(kwarg):
    """Int Keyword Argument."""

    regex = '[0-9]*'


class bool(kwarg):
    """Bool Keyword Argument."""

    regex = 'True|False'




# django_auto_url - Automagic URLs for Django
# Copyright (C) 2019 Thomas Hartmann <thomas.hartmann@th-ht.de>
#
# This file is part of django_auto_url.
#
# django_auto_url is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# django_auto_url is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with django_auto_url.  If not, see <http://www.gnu.org/licenses/>.
import inspect

import django.urls
from django_auto_url.utils.lazy import lazy

from django_auto_url.mixins import AutoUrlMixin
from django_auto_url.utils.class2urlname import url_cache, Class2URLName


def get_urls_from_module(module):
    """Scan a package and return the urlpatterns.

    This function scans the module for all subclasses of
    :class:`~django_auto_url.mixins.mixins.AutoUrlMixin`,  generates
    a :class:`~django.urls.resolvers.URLResolver` for each of them
    and returns a list that you can feed directly to :code:`urlpatterns`.

    Parameters
    ----------
    module : module
        The module to scan.

    Returns
    -------
    list of :class:`~django.urls.resolvers.URLResolver`

    Example
    -------
    Assuming, you have your views defined in the module
    :code:`my_app.views`, this is, how your :code:`urls.py` would look like:

    >>> from my_app import views
    >>> from django_auto_url.urls import get_urls_from_module
    >>>
    >>> urlpatterns = get_urls_from_module(views)

    """
    package_classes = [mod[1] for mod in
                       inspect.getmembers(module, inspect.isclass) if
                       mod[1].__module__ == module.__name__]
    url_classes = [cls for cls in package_classes if
                   issubclass(cls, AutoUrlMixin)]

    all_urls = []
    for cur_class in url_classes:
        url_cache['.'.join([inspect.getmodule(cur_class).__name__,
                            cur_class.__name__])] = cur_class
        rval = cur_class.get_url()
        if isinstance(rval, list):
            all_urls.extend(rval)
        else:
            all_urls.append(rval)

    return all_urls


def reverse_classname(class_name, return_url_name=False,
                      args=None, kwargs=None):  # noqa: D207, D301
    """Get the URL for a View.

    Converts a :class:`~django.views.generic.base.View` into its URL or
    the django url name.

    Parameters
    ----------
    class_name : str or subclass of \
:class:`~django_auto_url.mixins.mixins.AutoUrlMixin`
        The :class:`~django.views.generic.base.View` to reverse. You can either
        provide the class directly or a string with the full module path.
    return_url_name : bool, optional
        If set to :code:`True`, return the django url name and not the
        final URL.
    args : list, optional
        The positional arguments for the view.
    kwargs : dict, optional
        The keyword arguments for the view.

    Returns
    -------
    str
        The URL or the django url name.

    """
    url_name = Class2URLName.get_lookup_class_urlname(class_name)
    if return_url_name:
        if args or kwargs:
            raise RuntimeError('args and kwargs not '
                               'allowed if return_url_name=True.')
        return url_name
    else:
        return django.urls.reverse(url_name, args=args, kwargs=kwargs)


def reverse_classname_lazy(class_name, return_url_name=False,
                           args=None, kwargs=None):
    """Lazy version of :func:`~django_auto_url.urls.urls.reverse_classname`.

    Instead of doing the reverse instantly, it is done, when the result is
    actually needed. This will help you when you need to specify a URL
    in a class attribute. The non-lazy version would instantly try to find
    the URL for the view, even if the view has not been declared yet
    (for instance, because it is declared later in the same file or in a
    different one).

    Arguments and return values are exactly the same as
    :func:`~django_auto_url.urls.urls.reverse_classname`.
    """
    r_lazy = lazy(reverse_classname, str)
    return r_lazy(class_name, return_url_name, args, kwargs)


def reverse_local_classname(class_name,
                            return_url_name=False,
                            args=None,
                            kwargs=None,
                            lazy=True):  # noqa: D207, D301
    """Get the URL of a view declared in the same module.

    Basically does the same as
    :func:`~django_auto_url.urls.urls.reverse_classname` but you do not
    need to specify the full module path to the view. The class name is
    enough.

    Parameters
    ----------
    class_name : str or subclass of \
:class:`~django_auto_url.mixins.mixins.AutoUrlMixin`
        The :class:`~django.views.generic.base.View` to reverse. You can either
        provide the class directly or a string with the full module path.
    return_url_name : bool, optional
        If set to :code:`True`, return the django url name and not the
        final URL.
    args : list, optional
        The positional arguments for the view.
    kwargs : dict, optional
        The keyword arguments for the view.
    lazy : bool, optional
        Determines whether the evaluation should be lazy.

    Returns
    -------
    str
        The URL or the django url name.
    """
    cur_frame = inspect.currentframe()
    class_name = '.'.join(
        [cur_frame.f_back.f_locals['__module__'], class_name])

    if lazy:
        return reverse_classname_lazy(class_name, return_url_name,
                                      args, kwargs)
    else:
        return reverse_classname(class_name, return_url_name,
                                 args, kwargs)




# django_auto_url - Automagic URLs for Django
# Copyright (C) 2019 Thomas Hartmann <thomas.hartmann@th-ht.de>
#
# This file is part of django_auto_url.
#
# django_auto_url is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# django_auto_url is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with django_auto_url.  If not, see <http://www.gnu.org/licenses/>.

from django import template
from django.template import TemplateSyntaxError
from django.template.base import kwarg_re
from django.template.defaulttags import URLNode

from django_auto_url.utils.class2urlname import Class2URLName

register = template.Library()


@register.tag
def url_from_class(parser, token):
    """Template tag to resolve URL from class."""
    bits = token.split_contents()
    if len(bits) < 2:
        raise TemplateSyntaxError(
            "'%s' takes at least one argument, the name of a url()." % bits[0])

    tmp_name = "'%s'" % \
               (Class2URLName.get_lookup_class_urlname(bits[1][1:-1]),)
    viewname = parser.compile_filter(tmp_name)

    args = []
    kwargs = {}
    asvar = None
    bits = bits[2:]
    if len(bits) >= 2 and bits[-2] == 'as':
        asvar = bits[-1]
        bits = bits[:-2]

    if len(bits):
        for bit in bits:
            match = kwarg_re.match(bit)
            if not match:
                raise TemplateSyntaxError("Malformed arguments to url tag")
            name, value = match.groups()
            if name:
                kwargs[name] = parser.compile_filter(value)
            else:
                args.append(parser.compile_filter(value))

    if args:
        raise template.TemplateSyntaxError('Only keyword arguments are '
                                           'allowed when using the '
                                           '"url_from_class" template tag.')

    return URLNode(viewname, args, kwargs, asvar)
